# legacy.py - parse legacy (v0.4) Droit Databases
# Copyright 2019 - Jakob Stolze
#
# This file is part of python-droit (https://github.com/jarinox/python-droit)


import xml.etree.cElementTree as _ET
import xml.dom.minidom as _minidom

from . import models as _models


def isValidLine(ddaFileLine: str) -> bool:
	"""Checks whether a Droit Database rule is valid or e.g. a comment."""
	valid = True
	if not("->" in ddaFileLine) or (len(ddaFileLine.split("->")) != 2):
		valid = False
	if valid:
		if(ddaFileLine.split("->")[0] == "" or ddaFileLine.split("->")[1] == ""):
			valid = False
	if valid:
		lin = ddaFileLine.split("->")[0].split(":")
		for i in range(0, len(lin)):
			if(len(lin[i].split("!")) != 2):
				valid = False
	if valid:
		lout = ddaFileLine.split("->")[1].split(":")
		for i in range(0, len(lout)):
			if(len(lout[i].split("!")) != 2):
				valid = False
	return valid


def parseDDA(filename: str) -> list:
	"""Parses a Droit Database Script file to the legacy list standard"""
	ddaFile = open(filename, "r").read().split("\n")
	ddaData = []
	for i in range(0, len(ddaFile)):
		if(isValidLine(ddaFile[i])):
			ddaData.append(ddaFile[i].split("->"))

	
	for i in range(0, len(ddaData)):
		ddaData[i][0] = ddaData[i][0].split(":")
		ddaData[i][1] = ddaData[i][1].split(":")
		for j in range(0, len(ddaData[i][0])):
			ddaData[i][0][j] = ddaData[i][0][j].split("!")
		for j in range(0, len(ddaData[i][1])):
			ddaData[i][1][j] = ddaData[i][1][j].split("!")

	return ddaData


def parseDroitXML(filename: str):
	"""Parse a Droit XML Database"""
	tree = _ET.parse(filename)
	root = tree.getroot()
	rules = []
	for child in root:
		if(child.tag == "droitxml"):
			for rule in child:
				if(rule.tag == "rule"):
					inputrules = []
					outputrules = []
					for inrules in rule:
						if(inrules.tag == "input"):
							for inrule in inrules:
								children = []
								for inchild in inrule:
									if(inchild.text != None):
										children.append(inchild.text)
								dri = _models.DroitRuleInput(inrule.tag.upper(), inrule.attrib, children)
								inputrules.append(dri)
					for outrules in rule:
						if(outrules.tag == "output"):
							for outrule in outrules:
								children = []
								for outchild in outrule:
									if(outchild.text != None):
										children.append(outchild.text)
								dro = _models.DroitRuleOutput(outrule.tag.upper(), children)
								outputrules.append(dro)
					dr = _models.DroitRule(inputrules, outputrules)
					rules.append(dr)
	return rules


def writeDroitXML(dda, filename: str):
	"""Write a parse Droit Database to XML"""
	droitdb = _ET.Element("droitdb")
	droitxml = _ET.SubElement(droitdb, "droitxml")
	
	for rule in dda:
		r = _ET.SubElement(droitxml, "rule")
		inp = _ET.SubElement(r, "input")
		out = _ET.SubElement(r, "output")
		for inRule in rule.input:
			sub = _ET.SubElement(inp, inRule.tag, attrib=inRule.attrib)
			for child in inRule.children:
				_ET.SubElement(sub, "item").text = child
		for outRule in rule.output:
			sub = _ET.SubElement(out, outRule.tag)
			for child in outRule.children:
				_ET.SubElement(sub, "item").text = child
		
	
	tree = _ET.ElementTree(droitdb)
	tree.write(filename)


def prettifyXML(filename: str):
	dom = _minidom.parse(filename)
	pretty = dom.toprettyxml()
	open(filename, "w").write(pretty)


# python-droit - a simple library for creating bots
# Copyright 2020 Jakob Stolze <https://github.com/jarinox>
#
# Version 1.1.0:9 beta
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
# This file is part of python-droit (https://github.com/jarinox/python-droit)

import os as _os
import time as _time
import importlib as _importlib
import inspect as _inspect

from . import models
from . import legacy

from . import loader as _loader
from . import dumper as _dumper

from typing import List as _List



class Database:
    """A Droit Database that can be used to process rules."""
    def __init__(self, multiSession=False):
        self.rules = []
        self.info = models.DroitDatabaseInfo()
        self.plugins = []
        self.pluginReq = {}
        self.cache = models.DroitCache()
        self.history = models.DroitHistory()
        self.temp = {"plugins": {}}

        if(multiSession):
            self.sessions = models.DroitMultiSession()
        else:
            self.sessions = False

    def parseLegacy(self, filename: str):
        """Legacy parsing algorithm for Droit Database Script (.dda)"""
        self.rules = _loader.parseLegacy(filename, self.plugins)

    def parseScript(self, filename: str, plugins=True, append=False):
        """Parse a Droit Database Script file (.dda)"""
        newRules = []

        string = open(filename, "r").read()

        if(self.plugins and plugins):
            newRules = _loader.parseScriptString(string, plugins=self.plugins)
        else:
            newRules = _loader.parseScriptString(string)

        infos = _loader.parseScriptInfoString(string)
        self.info.add(infos)
        
        if(append):
            for rule in newRules:
                self.rules.append(rule)
        else:
            self.rules = newRules
    
    def parseScriptString(self, text: str, plugins=True, append=False):
        """Parse Droit Database Script from a string"""
        newRules = []

        if(self.plugins and plugins):
            newRules = _loader.parseScriptString(text, plugins=self.plugins)
        else:
            newRules = _loader.parseScriptString(text)
        
        infos = _loader.parseScriptInfoString(text)
        self.info.add(infos)
        
        if(append):
            for rule in newRules:
                self.rules.append(rule)
        else:
            self.rules = newRules
    
    def writeScript(self, filename: str):
        """Write a parsed Droit Database to a Droit Database Script file"""
        _dumper.writeScript(self.rules, filename)
    
    def writeScriptString(self) -> str:
        """Write a parsed Droit Database to a Droit Database Script string"""
        return _dumper.writeScriptString(self.rules)

    def loadPlugins(self, location=_os.path.join(_os.path.dirname(__file__), "plugins"), append=True, preloadScript=True):
        """
        Loads all plugins from the given location and returns them in a
        list containing DroitPlugin items.
        """
        if not(append):
            self.plugins = []

        inList = _os.listdir(path=_os.path.join(location, "input"))
        outList = _os.listdir(path=_os.path.join(location, "output"))
        
        for name in inList:
            if(_os.path.isdir(_os.path.join(location, "input", name))):
                plugin = models.DroitPlugin("input", name, path=location, preloadScript=preloadScript)
                if(plugin.info.req):
                    spec = _importlib.util.spec_from_file_location("pluginReq", _os.path.join(location, "input", name, "req.py"))
                    pl = _importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(pl)
                    members = _inspect.getmembers(pl, predicate=_inspect.isclass)
                    for member in members:
                        if(member[1].__name__ in self.pluginReq.keys()):
                            self.pluginReq[members][member[1].__name__].append(member[1])
                        else:
                            self.pluginReq[member[1].__name__] = [member[1]]

                self.plugins.append(plugin)
        
        for name in outList:
            if(_os.path.isdir(_os.path.join(location, "output", name))):
                self.plugins.append(models.DroitPlugin("output", name, path=location))
    
    def getPluginRequirements(self, plugin: str, func: str):
        """Load requirements provided by other plugins."""
        reqs = []
        for cl in self.pluginReq[plugin.lower()]:
            cl = cl()
            members = _inspect.getmembers(cl, predicate=_inspect.ismethod)
            for member in members:
                if(member[0] == func):
                    function = getattr(cl, func)
                    reqs.append(function)
                    break
        return reqs

    def createVariables(self, vars={}, userinput=None) -> dict:
        """
        Create a dict of variables containing all necessary pieces of data
        for formatOut
        """
        variables = {}
        variables["global.time"] = _time.strftime("%H:%M")
        variables["global.date"] = _time.strftime("%d.%m.%Y")
        variables["global.droitname"] = "Droit"
        variables["global.username"] = ""

        if(self.sessions):
            active = self.sessions.getActive()
            if(active):
                variables["global.username"] = active.username

                if(self.sessions.droitname):
                    variables["global.droitname"] = self.sessions.droitname
                else:
                    variables["global.droitname"] = active.droitname
            else:
                if(self.sessions.droitname):
                    variables["global.droitname"] = self.sessions.droitname

        
        if userinput:
            variables["global.userinput"] = userinput.rawInput
        else:
            variables["global.userinput"] = ""
        
        for var in vars:
            variables[var] = vars[var]
            
        return variables
    

    def useRules(self, userinput: models.DroitUserinput) -> _List[models.DroitSearchHit]:
        """
        Runs every rule onto the userinput.
        Returns all possible DroitRulesOutput sorted by relevance.
        """
        hits = []

        if not(self.plugins):
            print("Warning: no plugins loaded")

        for i in range(0, len(self.rules)): # use all rules
            variables = {}
            ranking = 0
            blocks = []
            ruleOk = True
            for j in range(0, len(self.rules[i].input)):
                if not(self.rules[i].input[j].tag in blocks):
                    blocks.append(self.rules[i].input[j].tag)
            
            for block in blocks:
                plug = None
                for inPlug in self.plugins:
                    if(inPlug.name == block.lower() and inPlug.mode == "input"):
                        plug = inPlug.plugin
                
                if(plug):
                    passRule, newVars, rankMod, self = plug.block(userinput, i, block, self)
                    for var in newVars:
                        variables[var] = newVars[var]
                    if(passRule):
                        ranking = ranking + rankMod
                    else:
                        ruleOk = False
                        break
                else:
                    ruleOk = False
            
            if(ruleOk):
                rankmod = len(self.rules[i].input) * 0.3 + ranking
                hit = models.DroitSearchHit(self.rules[i], variables, rankmod)
                hits.append(hit)

        if(hits != []):
            hits = sorted(hits, key=lambda hit: hit.ranking, reverse=True)
        
        return hits


    def formatOut(self, hit: models.DroitSearchHit, userinput: models.DroitUserinput) -> str:
        """Evaluates a DroitRuleOutput"""
        output = ""
        variables = self.createVariables(vars=hit.variables, userinput=userinput)

        if not(self.plugins):
            print("Warning: no plugins loaded")

        for i in range(0, len(hit.rule.output)):
            if(hit.rule.output[i].tag.upper() == "EVAL"):
                plugin = ""

                for plug in hit.rule.output[i].children:
                    plugin += plug
                plugin = plugin.split(".", 1)
                for var in variables:
                    if("*" + var in plugin[1]):
                        plugin[1] = plugin[1].replace("*" + var, "\"" + variables[var] + "\"")

                if("(" in plugin[1]):
                    plugin.append(plugin[1].split("(")[1][:-1].replace('"', "").replace(", ", ",").split(","))
                    plugin[1] = plugin[1].split("(")[0]

                params = []
                for i in range(0, len(plugin[2])):
                    params.append(plugin[2][i])
                
                plug = None
                for outPlug in self.plugins:
                    if(outPlug.name == "eval."+plugin[0] and outPlug.mode == "output"):
                        plug = outPlug.plugin
                
                if(plug):
                    method = getattr(plug, plugin[1])
                    
                    outadd = ""
                    outadd, self = method(params, self)

                    output += outadd
            
            
            else:
                plug = None
                for plugin in self.plugins:
                    if(plugin.mode == "output" and hit.rule.output[i].tag.lower() == plugin.name.lower()):
                        plug = plugin.plugin
                
                if(plug):
                    outadd = ""
                    outadd, self = plug.block(hit.rule.output[i], variables, self)
                    output += outadd
        
        return output
    

    def simpleIO(self, text: str, history=True) -> str:
        userinput = models.DroitUserinput(text)

        hits = self.useRules(userinput)
        if(hits):
            hit = hits[0]

            output = self.formatOut(hit, userinput)

            if(history):
                if(self.sessions):
                    if(self.sessions.getActive()):
                        self.history.newEntry(userinput, hit.rule, output, userId=self.sessions.getActive().id)
                    else:
                        self.history.newEntry(userinput, hit.rule, output)
                else:
                    self.history.newEntry(userinput, hit.rule, output)
            
            return output
        else:
            return None



# main.py - plugins.input - SRTX plugin for python-droit
# Copyright 2020 - Jakob Stolze
#
# This file is part of python-droit (https://github.com/jarinox/python-droit)

def block(userinput, iN, block, db):
	passRule = False
	variables = []
	rankMod = 2

	inputRules = db.rules[iN].input

	if(len(inputRules) == 1 and inputRules[0].tag == block):
		if(userinput.simpleInput in inputRules[0].children):
			passRule = True
	
	return passRule, variables, rankMod, db




# main.py - plugins.session - SESSION plugin for python-droit
# Copyright 2020 - Jakob Stolze
#
# This file is part of python-droit (https://github.com/jarinox/python-droit)
#
# SESSION*isActive!               is there an active session?
# SESSION*isNotActive!            is there no active session?
# SESSION*isUsername!<username>   = active username?
# SESSION*isId!<id>               = active id?


def block(userinput, iN, name, db):
	passRule = False
	rankMod = 0
	outVars = {}

	inputRules = db.rules[iN].input

	for rule in inputRules:
		if(rule.tag == "SESSION"):
			if(rule.attrib["cmd"] == "isActive"): # isActive
				if(db.sessions):
					if(db.sessions.getActive()):
						passRule = True
			
			if(rule.attrib["cmd"] == "isNotActive"): # isNotActive
				if not(db.sessions):
					passRule = True
				elif not(db.sessions.getActive()):
					passRule = True
			
			if(rule.attrib["cmd"] == "isUsername"): # isUsername
				if(db.sessions):
					if(db.sessions.getActive()):
						if(db.sessions.getActive().username in rule.children):
							passRule = True
			
			if(rule.attrib["cmd"] == "isId"): # isId
				if(db.sessions):
					if(db.sessions.getActive()):
						if(db.sessions.getActive().id in rule.children):
							passRule = True

	return passRule, outVars, rankMod, db


# main.py - plugins.input - SIMT (similar text) plugin for python-droit
# Copyright 2020 - Jakob Stolze
#
# This file is part of python-droit (https://github.com/jarinox/python-droit)

import difflib

def block(userinput, iN, block, db):
    passRule = False
    variables = []
    rankMod = 0

    inputRules = db.rules[iN].input

    for j in range(0, len(inputRules)):
        if(block == inputRules[j].tag):
            limit = 0.9
            if("limit" in inputRules[j].attrib.keys()):
                try:
                    limit = float(inputRules[j].attrib["limit"]) / 100
                except:
                    print("python-droit, error: plugins.input.simt, parsing 'limit' attrib failed")

            for child in inputRules[j].children:
                difference = difflib.SequenceMatcher(None, child, userinput.simpleInput).ratio()
                if(difference >= limit):
                    passRule = True

    return passRule, variables, rankMod, db



# main.py - plugins.input - TEXT plugin for python-droit
# Copyright 2020 - Jakob Stolze
#
# This file is part of python-droit (https://github.com/jarinox/python-droit)

def block(userinput, iN, block, db):
	passRule = True
	variables = []
	rankMod = 1

	inputRules = db.rules[iN].input
	
	for j in range(0, len(inputRules)):
		if(block == inputRules[j].tag):
			dont = False
			if "not" in inputRules[j].attrib:
				dont = (inputRules[j].attrib["not"] == "true")
				if(dont):
					rankMod = 2
			
			thisRule = False
			for i in range(0, len(inputRules[j].children)):
				if(inputRules[j].children[i].split(" ") == 1):
					if(inputRules[j].children[i] in userinput.words and not(dont)):
						thisRule = True
					if(not(inputRules[j].children[i] in userinput.words) and dont):
						thisRule = True
				else:
					if(inputRules[j].children[i] in userinput.simpleInput and not(dont)):
						thisRule = True
					if(not(inputRules[j].children[i] in userinput.simpleInput) and dont):
						thisRule = True
			
			if not(thisRule):
				passRule = False
	
	return passRule, variables, rankMod, db



# main.py - plugins.input - INP plugin for python-droit
# Copyright 2020 - Jakob Stolze
#
# This file is part of python-droit (https://github.com/jarinox/python-droit)

from parse import parse


def block(userinput, iN, name, db):
	pTexts = []
	passRule = False
	rankMod = 0
	varChildren = {}

	inputRules = db.rules[iN].input
	reqs = db.getPluginRequirements(name, "placeholder")

	for rule in inputRules:
		for function in reqs:
			pTexts = function(rule, pTexts)
		
		if(rule.tag == name.upper()):
			varname = rule.attrib["var"]
			if(len(rule.children) > 0):
				varname = "strict." + varname
				varChildren[varname] = rule.children
			if(pTexts == []):
				pTexts.append("{" + varname + "} ")
			else:
				pcop = pTexts
				for k in range(0, len(pcop)):
					pTexts[k] = pTexts[k] + "{" + varname + "} "
	
	for i in range(0, len(pTexts)):
		if(pTexts[i][-1] == " "):
			pTexts[i] = pTexts[i][0:len(pTexts[i])-1]

	variables = {}
	for pText in pTexts:
		results = parse(pText, userinput.rawInput)
		if(results != None):
			passRule = True
			
			variables = results.__dict__["named"]
			for variable in variables:
				if("strict." in variable):
					value = variables[variable]
					if not(value in varChildren[variable]):
						passRule = False
	
	outVars = {}
	for variable in variables:
		outVars["inp."+variable.replace("strict.", "")] = variables[variable]

	return passRule, outVars, rankMod, db


# main.py - plugins.output - TEXT plugin for python-droit
# Copyright 2020 - Jakob Stolze
#
# This file is part of python-droit (https://github.com/jarinox/python-droit)

def block(rule, variables, db):
    output = ""
    for item in rule.children:
        output += item + ","
    output = output[:-1]
    return output, db


# main.py - plugins.output - EVAL.session plugin for python-droit
# Copyright 2020 - Jakob Stolze
#
# This file is part of python-droit (https://github.com/jarinox/python-droit)


def activateByUsername(data, db):
    if(db.sessions):
        db.sessions.activateByUsername(data[0])
    return "", db

def activateById(data, db):
    if(db.sessions):
        db.sessions.activateById(data[0])
    return "", db

def saveSessions(data, db):
    if(db.sessions):
        db.sessions.saveSessions()
    return "", db

def getUsername(data, db):
    if(db.sessions):
        if(db.sessions.getActive()):
            return db.sessions.getActive().username, db
        else:
            return "", db
    else:
        return "", db

def setDroitname(data, db):
    db.sessions.droitname = data[0]
    return "", db

def setUsername(data, db):
    try:
        active = db.sessions.getActive()
        active.username = data[0]
        active.userData["customUsername"] = True
        db.session.setActive(active)
    except:
        pass
    return "", db


# main.py - plugins.output - VAR plugin for python-droit
# Copyright 2020 - Jakob Stolze
#
# This file is part of python-droit (https://github.com/jarinox/python-droit)

def block(rule, variables, db):
    return variables[rule.children[0]], db


from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
	name='droit',
	version='1.1.0.beta.9',
	description='Simple library for creating bots',
	long_description=long_description,
    long_description_content_type="text/markdown",
	url='https://github.com/jarinox/python-droit',
	author='Jakob Stolze',
	author_email='jarinox@wolke7.de',
	license='LGPLv2.1',
	packages=['droit'],
	install_requires=['parse'],
	include_package_data=True,
	python_requires='>=3.6',
	zip_safe=False
)



# models.py - structure python-droit data
# Copyright 2020 - Jakob Stolze
#
# This file is part of python-droit (https://github.com/jarinox/python-droit)


import os as _os
import importlib as _importlib
import json as _json
import random as _random



class DroitSession:
	def __init__(self, username: str, droitname=False, ident=_random.randint(0, 100000000000000)):
		self.username = username
		self.droitname = droitname
		self.userData = {}
		self.id = ident

	def fromDict(self, var: dict):
		self.id = var["id"]
		self.username = var["username"]
		self.droitname = var["droitname"]
		self.userData = var["userData"]
	
	def toDict(self) -> dict:
		return {
			"id": self.id,
			"username": self.username,
			"droitname": self.droitname,
			"userData": self.userData
		}

class DroitMultiSession:
	def __init__(self, path=None, droitname=False):
		self.sessions = []
		self.active = -1
		self.path = None
		self.droitname=droitname
	
	def activateByUsername(self, username):
		for i in range(0, len(self.sessions)):
			if(self.sessions[i].username == username):
				self.active = i
				return True
		return False
	
	def activateById(self, ident):
		for i in range(0, len(self.sessions)):
			if(self.sessions[i].id == ident):
				self.active = i
				return True
		return False
	
	def getActive(self) -> DroitSession:
		if(self.active != -1 and self.active < len(self.sessions)):
			return self.sessions[self.active]
		else:
			return False
	
	def setActive(self, session):
		if(self.active != -1 and self.active < len(self.sessions)):
			self.sessions[self.active] = session
	
	def loadSessions(self, append=False):
		if not(append):
			self.sessions = []
		
		with open(self.path, "r") as f:
			data = _json.load(f)
			self.droitname = data["droitname"]
			self.active = data["active"]

			sessions = data["sessions"]
			for session in sessions:
				s = DroitSession("")
				s.fromDict(session)
				self.sessions.append(s)

	def saveSessions(self):
		dump = []
		for session in self.sessions:
			dump.append(session.toDict())
		
		with open(self.path, "w") as f:
			_json.dump({"sessions": dump, "droitname": self.droitname, "active": self.active}, f)

class DroitCache:
	"""Cache the return value of slow functions"""
	def __init__(self):
		self.storage = []
	
	def run(self, function, param1=None, param2=None, param3=None):
		name = function.__module__ +  "." + function.__name__
		params = [param1, param2, param3]

		for item in self.storage:
			if(item["name"] == name and item["params"] == params):
				return item["value"]
			
		value = None
		if(param1 == None):
			value = function()
		elif(param2 == None):
			value = function(param1)
		elif(param3 == None):
			value = function(param1, param2)
		else:
			value = function(param1, param2, param3)
		
		self.storage.append({"name": name, "params": params, "value": value})
		return value


class DroitRuleInput:
	"""Stores an input-rule"""
	def __init__(self, tag: str, attrib: dict, children: list):
		self.mode = "input"
		self.tag = tag.upper()
		self.attrib = attrib
		self.children = children


class DroitRuleOutput:
	"""Stores an output-rule"""
	def __init__(self, tag: str, children: list):
		self.mode = "input"
		self.tag = tag.upper()
		self.children = children


class DroitRule:
	"""Stores a list of inputRules and a list of outputRules."""
	def __init__(self, inputRules: list, outputRules: list):
		self.input = inputRules
		self.output = outputRules


class DroitUserinput:
	"""
	Stores the raw userinput as well as a list of the words the userinput
	consists of. The list is created on init.
	"""
	def __init__(self, rawInput: str):
		self.rawInput = rawInput
		pin = ""
		valchars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZäöüÄÖÜ1234567890ß "
		for char in rawInput: # remove unnecessary characters
			if(char in valchars):
				pin += char

		while("  " in pin):
			pin = pin.replace("  ", " ") # double blank to single blank
		
		pin = pin.rstrip()
		self.caseInput = pin
		self.simpleInput = pin.lower()
		self.words = self.simpleInput.split(" ") # split up words at blank


class DroitPlugin:
	"""Loads a plugin."""
	def __init__(self, mode: str, name: str, path=_os.path.join(_os.path.dirname(__file__), "plugins"), preloadScript=True):
		self.mode = mode.lower()
		self.name = name.lower()
		spec = _importlib.util.spec_from_file_location("main", _os.path.join(path, mode, name, "main.py"))
		self.plugin = _importlib.util.module_from_spec(spec)
		spec.loader.exec_module(self.plugin)
		self.info = DroitPluginInfo(mode, name, path=path)
		if(self.mode == "input"):
			if("preloadScript" in self.info.info.keys() and preloadScript):
				_os.system("python3 " + _os.path.join(path, "input", name, self.info.info["preloadScript"]))


class DroitPluginInfo:
	"""Contains information about a DroitPlugin"""
	def __init__(self, mode: str, name: str, path=_os.path.join(_os.path.dirname(__file__), "plugins")):
		self.mode = mode
		self.name = name
		if(mode == "input"):
			self.info = _json.loads(open(_os.path.join(path, mode, name, "info.json"), "r").read())
			self.description = self.info["description"]
			self.attrib = self.info["attributes"]
			files = _os.listdir(_os.path.join(path, mode, name))
			self.req = ("req.py" in files)


class DroitSearchHit:
	"""Object returned by useRules()"""
	def __init__(self, rule: DroitRule, variables: dict, ranking: int):
		self.rule = rule
		self.variables = variables
		self.ranking = ranking


class DroitHistory:
	"""List of inputs and outputs of python-droit."""
	def __init__(self):
		self.inputs = []
		self.outputs = []
		self.rules = []
		self.userIds = []

	def newEntry(self, userinput: DroitUserinput, rule: DroitRule, output: str, userId=False):
		self.inputs.append(userinput)
		self.rules.append(rule)
		self.outputs.append(output)
		self.userIds.append(userId)


class DroitDatabaseInfo:
	"Information about a Droit Datbase"
	def __init__(self, author=None, license=None, ddsVersion=None, attrib={}):
		if(author):
			self.author = [author]
		else:
			self.author = []
		
		if(ddsVersion):
			self.ddsVersion = [ddsVersion]
		else:
			self.ddsVersion = []

		if(license):
			self.license = [license]
		else:
			self.license = []

		self.attrib = attrib


	def add(self, attribs):
		for key in attribs:
			lkey = key.lower()
			if(lkey == "author"):
				if not(attribs[key] in self.author):
					self.author.append(attribs[key])	
			elif(lkey =="license" or lkey == "licence"):
				if(not license in self.license):
					self.license.append(attribs[key])		
			elif(lkey == "dds" or lkey == "ddsversion"):
				if not(attribs[key] in self.ddsVersion):
					self.ddsVersion.append(attribs[key])
			else:
				self.attrib[lkey] = attribs[key]


class inp:
    def __init__(self):
        self.tag = "TEXT"
    
    def placeholder(self, rule, pTexts):
        if(rule.tag == self.tag):
            if(pTexts == []):
                for child in rule.children:
                    pTexts.append(child + " ")
            else:
                pNew = []
                for i in range(0, len(pTexts)):
                    for child in rule.children:
                        pNew.append(pTexts[i] + child + " ")
                    pTexts = pNew
        
        return pTexts



# loader.py - parse Droit Databases
# Copyright 2020 - Jakob Stolze
#
# This file is part of python-droit (https://github.com/jarinox/python-droit)



from . import models as _models
from . import legacy as _legacy

from typing import List as _List


def parseLegacy(filename: str, plugins) -> _List[_models.DroitRule]:
	"""Parse a legacy Droit Database (.dda)"""
	dda = _legacy.parseDDA(filename)
	rules = []
	pinfos = []
	for plugin in plugins:
		pinfos.append(plugin.info)
	for rule in dda:
		inputrules = []
		outputrules = []
		for inrule in rule[0]:
			attr = {}
			inrule[1] = inrule[1].replace("&arz;", "!").replace("&dpp;", ":")
			if("*" in inrule[0]):
				for info in pinfos:
					if(info.name.lower() == inrule[0].split("*")[0].lower()):
						ipkey = list(info.attrib.keys())[0]
						attr[ipkey] = inrule[0].split("*")[1]
				inrule[0] = inrule[0].split("*")[0]

			if("NOTX" == inrule[0]):
				inrule[0] = "TEXT"
				attr["not"] = "true"
			children = inrule[1].split(",")
			if(inrule[1] == ""):
				children = []
			dri = _models.DroitRuleInput(inrule[0], attr, children)
			inputrules.append(dri)
		for outrule in rule[1]:
			outrule[1] = outrule[1].replace("&arz;", "!").replace("&dpp;", ":")
			if(outrule[0].upper() == "EVAL"):
				dro = _models.DroitRuleOutput(outrule[0].upper(), [outrule[1]])
			else:
				children = outrule[1].split(",")
				if(outrule[1] == ""):
					children = []
				dro = _models.DroitRuleOutput(outrule[0].upper(), children)
			outputrules.append(dro)
		dr = _models.DroitRule(inputrules, outputrules)
		rules.append(dr)
	return rules
		

def parseScript(filename: str, plugins=False) -> _List[_models.DroitRule]:
	plain = open(filename, "r").read()
	return parseScriptString(plain, plugins=plugins)

def parseScriptString(string: str, plugins=False) -> _List[_models.DroitRule]:
	plain = string.split("\n")
	rules = []
	
	for line in plain:
		if(_legacy.isValidLine(line)):
			rule = _models.DroitRule([], [])
			inp, out = line.split("->")

			inp = inp.split(":")
			out = out.split(":")

			for inpx in inp:
				inpx = inpx.split("!")
				inpx[0] = inpx[0].replace("NOTX", "TEXT*true")
				
				if(inpx[1] != ""):
					children = inpx[1].replace("&arz;", "!").replace("&dpp;", ":").split(",")
				else:
					children = []

				if("*" in inpx[0]):
					attr = {}
					block, atnm = inpx[0].split("*")
					block = block.upper()

					if(plugins):
						for plugin in plugins:
							if(plugin.mode == "input" and plugin.name.upper() == block):
								attr[list(plugin.info.attrib.keys())[0]] = atnm
					else:
						if(block == "INP"):
							attr["var"] = atnm
						elif(block == "SIMT"):
							attr["limit"] = atnm
						elif(block == "TEXT"):
							attr["not"] = atnm
						elif(block == "SESSION"):
							attr["cmd"] = atnm
					
					rule.input.append(_models.DroitRuleInput(block, attr, children))
				else:
					rule.input.append(_models.DroitRuleInput(inpx[0], {}, children))
			
			for outx in out:
				outx = outx.split("!")
				if(outx[1] != ""):
					if(outx[0].upper() == "EVAL"):
						children = [outx[1].replace("&arz;", "!").replace("&dpp;", ":")]
					else:
						children = outx[1].replace("&arz;", "!").replace("&dpp;", ":").split(",")
				else:
					children = []
				rule.output.append(_models.DroitRuleOutput(outx[0], children))

			rules.append(rule)	

	return rules


def parseScriptInfoString(string: str) -> dict:
	data = string.split("\n")
	ret = {}
	for line in data:
		if(len(line) > 3):
			if(line[0] == "@" and " " in line):
				parts = line.split(" ")
				if(len(parts) > 0):
					if(" ".join(parts[1:]).replace(" ", "") != ""):
						ret[parts[0][1:]] = " ".join(parts[1:])
	
	return ret


# dumper.py - write a parsed Droit Databases to XML
# Copyright 2020 - Jakob Stolze
#
# This file is part of python-droit (https://github.com/jarinox/python-droit)



def writeScript(dda, filename):
	"""Write a parse Droit Database to a Droit Database Script file"""
	out = writeScriptString(dda)
	open(filename, "w").write(out)

def writeScriptString(dda) -> str:
	"""Write a parse Droit Database to a Droit Database Script string"""
	out = "Created using python-droit and 'writeScript'\n\n"

	for rule in dda:
		for inRule in rule.input:
			out += inRule.tag.upper()

			if(inRule.attrib):
				out += "*" + list(inRule.attrib.items())[0][1]
			out += "!"

			for value in inRule.children:
				if(out[-1] != "!"):
					out += ","
				out += value.replace(":", "&dpp;").replace("!", "&arz;")
			
			out += ":"
		
		out = out[0:-1] + "->"

		for outRule in rule.output:
			out += outRule.tag.upper()
			out += "!"
			
			for value in outRule.children:
				if(out[-1] != "!"):
					out += ","
				out += value.replace(":", "&dpp;").replace("!", "&arz;")
			
			out += ":"
		
		out = out[0:-1] + "\n"
	
	out = out.replace("TEXT*true!", "NOTX!")
	return out
